"""
Classes for enforcing correlations between branches in logic trees
"""
from dataclasses import dataclass, field
from typing import Generator, List, Optional, Sequence, overload

from .branch import Branch


@dataclass
class Correlation:
    """
    A correlation between **BranchSet**s of a logic tree. Correlations are used when buiding combinations of branches
    from differnt **BranchSet**s.

    For example, if a logic tree contains branchets A and B each with branches A1, A2 and B1, B2, respectivly, then
    a correlation A1, B1 will only allow A1-B1 not A1-B2 as valid composite branches.

    The primary_brach will not appear in combination with any branches except those identified in associated_branches
    for the relevent **BranchSet**s

    Arguments:
        primary_branch: the branch that **MUST** be correlated with the associated brances.
        associated_branches: list of branches that the primary_branch must always be coupled with.
        weight: weight used for composite branch formed by correlation. Defaults to weight of primary_branch
    """

    primary_branch: Branch = field(default_factory=Branch)
    associated_branches: List[Branch] = field(default_factory=list)
    weight: Optional[float] = None

    def __post_init__(self):
        self.weight = self.primary_branch.weight if not self.weight else self.weight

    @property
    def all_branches(self) -> List[Branch]:
        """
        list of all branches in a correlation; both the primary branch and the assiciated branches

        Returns:
            list of branches in correlation
        """
        return [self.primary_branch] + self.associated_branches


@dataclass(frozen=True)
class LogicTreeCorrelations(Sequence):
    """
    All correlations for a logic tree.

    Arguments:
        correlation_groups: list of correlations to be applied to the logic tree branch sets.
    """

    correlation_groups: List[Correlation] = field(default_factory=list)

    def __post_init__(self) -> None:
        _validate_correlations(self)

    def primary_branches(self) -> Generator[Branch, None, None]:
        """
        Yield the primary branches of all correlation_groups

        Returns:
            branches
        """
        for cor in self.correlation_groups:
            yield cor.primary_branch

    @overload
    def __getitem__(self, i: int) -> Correlation:
        ...

    @overload
    def __getitem__(self, i: slice) -> Sequence[Correlation]:
        ...

    def __getitem__(self, i):
        if isinstance(i, slice):
            raise TypeError("LogicTreeCorrelations does not support slicing")
        return self.correlation_groups[i]

    def __len__(self) -> int:
        return len(self.correlation_groups)


def _validate_correlations(ltcs: LogicTreeCorrelations) -> None:
    """
    check that there are no repeats in the 0th element of each correlation
    """
    prim_branches = list(ltcs.primary_branches())
    if len([branch for branch in prim_branches if prim_branches.count(branch) > 1]) != 0:
        raise ValueError("there is a repeated branch in the 0th element of the correlations")



"""
The logic_tree package provides dataclasses that define the different types of LogicTree
and Logic Tree Branches

Classes:
    LogicTree: an Abstract Base Class for LogicTrees
    SourceLogicTree: dataclass for source logic trees
    SourceLogicTreeV1: (Deprecated)
    GMCMLogicTree: dataclass for Ground Motion Model logic trees
    SourceBranch: dataclass for individual source branches
    SourceBranchSet: dataclass for grouping source branches

"""

from .branch import CompositeBranch
from .correlation import Correlation, LogicTreeCorrelations
from .gmcm_logic_tree import GMCMBranch, GMCMBranchSet, GMCMLogicTree
from .source_logic_tree import InversionSource, SourceBranch, SourceBranchSet, SourceLogicTree, SourceLogicTreeV1



"""
NshmModel class describes a complete National Seismic Hazard Model.
"""
import json
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, List, Union, cast

from nzshm_model.logic_tree import GMCMLogicTree, SourceBranchSet, SourceLogicTree
from nzshm_model.logic_tree.source_logic_tree import SourceLogicTreeV1
from nzshm_model.psha_adapter.openquake import NrmlDocument, OpenquakeSimplePshaAdapter

if TYPE_CHECKING:
    from nzshm_model import psha_adapter

RESOURCES_PATH = Path(__file__).parent.parent / "resources"
SLT_SOURCE_PATH = RESOURCES_PATH / "SRM_JSON"
GMM_SOURCE_PATH = RESOURCES_PATH / "GMM_LTs"

versions = {
    "NSHM_v1.0.0": 'nzshm_model.nshm_v1_0_0',
    "NSHM_v1.0.4": 'nzshm_model.nshm_v1_0_4',
}


# https://stackoverflow.com/questions/48976499/mypy-importlib-module-functions
class ModuleInterface:
    """a type interface for this module

    it doesn't have to be instantiated, it'll just help mypy figure things out)
    """

    model: 'NshmModel'


def import_module_with_interface(modname: str) -> ModuleInterface:
    return __import__(modname, fromlist=['_trash'])  # type: ignore


class NshmModel:
    """
    An NshmModel instance represents a complete National Seismic Hazard Model version.
    """

    def __init__(self, version, title, slt_json, gmm_xml, slt_config):
        """
        Create a new NshmModel instance.

        NB library users will typically never use this, rather they will obtain a model instance
        using static method: `get_model_version`.
        """
        self.version = version
        self.title = title
        self.slt_config = slt_config

        self._slt_json = SLT_SOURCE_PATH / slt_json
        self._gmm_xml = GMM_SOURCE_PATH / gmm_xml
        assert self._slt_json.exists()
        assert self._gmm_xml.exists()

    @property
    def _data(self):
        with open(self._slt_json, 'r') as jsonfile:
            data = json.load(jsonfile)
        return data

    @property
    def source_logic_tree(self) -> "SourceLogicTree":
        """
        the source logic tree for this model.

        Returns:
            a source_logic_tree

        """
        data = self._data
        ltv = data.get("logic_tree_version")
        if ltv is None:  # original json is unversioned
            return SourceLogicTree.from_source_logic_tree(SourceLogicTreeV1.from_dict(data))
        raise ValueError("Unsupported logic_tree_version.")

    def source_logic_tree_nrml(self) -> "psha_adapter.openquake.logic_tree.LogicTree":
        """
        the Source logic tree for this model as a OpenQuake nrml compatiable type.

        Returns:
            a source_logic_tree
        """
        warnings.warn("use NshmModel.source_logic_tree().psha_adapter().config() instead", DeprecationWarning)
        slt = self.source_logic_tree
        return slt.psha_adapter(provider=OpenquakeSimplePshaAdapter).config()

    @property
    def gmm_logic_tree(self) -> "GMCMLogicTree":
        """
        the ground motion logic tree for this model.

        Returns:
            a gmcm_logic_tree

        """
        adapter = GMCMLogicTree().psha_adapter(OpenquakeSimplePshaAdapter)
        return adapter.logic_tree_from_xml(self._gmm_xml)  # type: ignore

    def gmm_logic_tree_nrml(self) -> "psha_adapter.openquake.logic_tree.LogicTree":
        """
        the ground motion characterisation model (gmcm) logic tree for this model as a OpenQuake nrml compatiable type.

        Returns:
            a gmm_logic_tree.
        """
        doc = NrmlDocument.from_xml_file(self._gmm_xml)
        return doc.logic_trees[0]

    @staticmethod
    def get_model_version(version: str) -> 'NshmModel':
        """
        Retrieve an existing model by its specific version

        Examples:
            >>> from nzshm_model import NshmModel
            >>> model = NshmModel.get_model_version("NSHM_v1.0.4")
            >>> print(model.title)
            >>>
            NSHM version 1.0.4, corrected fault geometry

        Parameters:
            version: The unique identifier for the model version.

        Raises:
            ValueError: when the version does not exist.

        Returns:
            the model instance.
        """
        model_spec_module = versions.get(version)
        if not model_spec_module:
            raise ValueError(f"{version} is not a valid model version.")
        # module = importlib.import_module(model_spec_module)
        module = import_module_with_interface(model_spec_module)
        # return module.model
        return cast('NshmModel', module.model)
        # return cast('NshmModel', model)

    def get_source_branch_sets(self, short_names: Union[List[str], str, None] = None) -> Iterator['SourceBranchSet']:
        """
        get an iterator for the SourceBranchSets matching the specified branch set(s)

        Examples:
            >>>  model = get_model_version("NSHM_v1.0.4")
            >>>  for branch_set in model.get_source_branch_sets(['CRU', 'PUY']):
                    print(branch_set.short_name, len(branch_set.branches))
            >>>
            CRU 36
            PUY 3

        Parameters:
            short_names: list of short_names for branch_set(s) (eg. 'HIK', 'CRU', 'PUY', 'SLAB')

        Raises:
            ValueError: when a branch short_name is not found.

        Yields:
            iterator of branch_set objects
        """
        if isinstance(short_names, str):
            list_short_names: List[str] = [short_names]
        else:
            list_short_names = short_names if short_names is not None else []

        if not list_short_names:  # User passes either an empty list or None
            for branch_set in self.source_logic_tree.branch_sets:
                yield branch_set
        else:
            # user has passes a list of short_names
            # check all the names are valid:
            for short_name in list_short_names:
                try:
                    for b in filter(lambda item: item.short_name == short_name, self.source_logic_tree.branch_sets):
                        yield (b)
                except StopIteration:
                    raise ValueError("The branch " + short_name + " was not found.")

    # def get_source_branches(self, short_names: list = None) -> Iterator['SourceBranch']:
    #     """
    #     get an iterator for the SourceBranches matching the specified branch set(s)

    #     Examples:
    #         >>>  model = get_model_version("NSHM_v1.0.4")
    #         >>>  for branch in model.get_source_branches(['CRU', 'PUY']):
    #                 print(branch.tag, branch.weight)
    #         >>>
    #         [dm0.7, bN[0.902, 4.6], C4.0, s0.28] 0.21
    #         ...

    #     Parameters:
    #         short_names: list of short_names for branch_set(s) (eg. HIK, CRU, PUY, SLAB)

    #     Raises:
    #         ValueError: when a branch short_name is not found.

    #     Yields:
    #         iterator of branch objects
    #     """
    #     for branch_set in self.get_source_branch_sets(short_names):
    #         for branch in branch_set.branches:
    #             yield (branch)



import pathlib
from abc import ABC, abstractmethod
from typing import Dict, Union


class PshaAdapterInterface(ABC):
    """
    Defines methods to be provided by a PSHA adapter class implementation.
    """

    def __init__(self, source_logic_tree=None, gmcm_logic_tree=None):
        self._source_logic_tree = source_logic_tree
        self._gmcm_logic_tree = gmcm_logic_tree

    @abstractmethod
    def fetch_resources(self, cache_folder):
        """pull required data from api and store in target_folder"""
        pass

    @abstractmethod
    def unpack_resources(self, cache_folder: Union[pathlib.Path, str], target_folder: Union[pathlib.Path, str]):
        """unpack resources from cache_folder to target folder returning mapping of unpacked filepaths"""
        pass

    @abstractmethod
    def config(self):
        """Get the PSHA config file"""
        pass

    @abstractmethod
    def write_config(
        self,
        cache_folder: Union[pathlib.Path, str],
        target_folder: Union[pathlib.Path, str],
        resource_map: Dict[str, list[pathlib.Path]],
    ) -> pathlib.Path:
        """Build an openquake config from given input arguments"""
        pass



#! logic_tree.py

"""
Define source logic tree structures used in NSHM.
"""
import json
import pathlib
from copy import deepcopy
from dataclasses import dataclass, field
from functools import reduce
from itertools import product
from math import isclose
from operator import add, mul
from typing import Dict, Iterable, List, Type, Union

import dacite

from nzshm_model.psha_adapter import PshaAdapterInterface

from .. import BranchAttributeValue
from ..core import BranchSetBase, BranchSetSpec


@dataclass
class Branch:
    values: List[BranchAttributeValue]
    weight: float = 1.0
    onfault_nrml_id: Union[str, None] = ""
    distributed_nrml_id: Union[str, None] = ""
    inversion_solution_id: Union[str, None] = ""
    inversion_solution_type: Union[str, None] = ""
    rupture_set_id: Union[str, None] = ""


@dataclass
class FaultSystemLogicTree(BranchSetBase):
    branches: List[Branch] = field(default_factory=list)


@dataclass
class SourceLogicTreeCorrelation:
    primary_short_name: str
    secondary_short_name: str
    primary_values: List[BranchAttributeValue]
    secondary_values: List[BranchAttributeValue]

    # these methods enforce set compairson
    def is_primary(self, bavs: Iterable[BranchAttributeValue]) -> bool:
        return all([item in self.primary_values for item in bavs])

    def is_secondary(self, bavs: Iterable[BranchAttributeValue]) -> bool:
        return all([item in self.secondary_values for item in bavs])


@dataclass
class SourceLogicTreeSpec:
    fault_system_lts: List[BranchSetSpec] = field(default_factory=list)


@dataclass
class SourceLogicTree:
    version: str
    title: str
    fault_system_lts: List[FaultSystemLogicTree] = field(default_factory=list)
    correlations: List[SourceLogicTreeCorrelation] = field(
        default_factory=list
    )  # to use for selecting branches and re-weighting when logic trees are correlated

    @property
    def fault_systems(self):
        return self.fault_system_lts

    def derive_spec(self) -> SourceLogicTreeSpec:
        slt_spec = SourceLogicTreeSpec()
        for fslt in self.fault_system_lts:
            slt_spec.fault_system_lts.append(FaultSystemLogicTree.derive_spec(fslt))
        return slt_spec

    def psha_adapter(self, provider: Type[PshaAdapterInterface], **kwargs):
        return provider(source_logic_tree=self)

    @staticmethod
    def from_dict(data: Dict):
        ltv = data.get("logic_tree_version")
        if ltv:
            raise ValueError(f"supplied json `logic_tree_version={ltv}` is not supported.")
        return dacite.from_dict(data_class=SourceLogicTree, data=data, config=dacite.Config(strict=True))

    @staticmethod
    def from_json(json_path: Union[pathlib.Path, str]):
        data = json.load(open(json_path))
        return SourceLogicTree.from_dict(data)


@dataclass
class CompositeBranch:
    """Combination of all fault type branches"""

    branches: List[Branch]
    weight: float = 1.0

    def __post_init__(self) -> None:
        self.weight = reduce(mul, [branch.weight for branch in self.branches])


@dataclass
class FlattenedSourceLogicTree:
    """Flattened source logic tree containing all CompositeBranch combinations"""

    version: str
    title: str
    branches: List[CompositeBranch]

    def __post_init__(self) -> None:
        total_weight = reduce(add, [branch.weight for branch in self.branches])
        if not isclose(total_weight, 1.0):
            raise Exception('logic tree weights do not add to 1.0 (sum is %s)' % total_weight)

    @classmethod
    def from_source_logic_tree(cls, slt: SourceLogicTree):

        slt_copy = deepcopy(slt)  # don't want to change weights of origional logic tree object
        branches = [fslt.branches for fslt in slt_copy.fault_system_lts]

        def yield_cor(branches, slt_copy):
            nnames = [[faultsys_lt.short_name] * len(faultsys_lt.branches) for faultsys_lt in slt_copy.fault_system_lts]
            for cb, names in zip(product(*branches), product(*nnames)):
                has_correlation = False
                for cor in slt_copy.correlations:
                    sindex = names.index(cor.secondary_short_name)
                    if cor.is_secondary(cb[sindex].values):
                        has_correlation = True
                        pindex = names.index(cor.primary_short_name)
                        if cor.is_primary(cb[pindex].values):
                            cb[sindex].weight = 1.0
                            yield CompositeBranch(list(cb))
                if not has_correlation:
                    yield CompositeBranch(list(cb))

        if slt.correlations:
            return cls(slt.version, slt.title, list(yield_cor(branches, slt_copy)))
        else:
            return cls(slt.version, slt.title, [CompositeBranch(list(cb)) for cb in product(*branches)])

    def __repr__(self):
        return f'{self.__class__} title {self.title} number of branches: {len(self.branches)}'



from pathlib import Path

import pytest

from nzshm_model.logic_tree import SourceLogicTree

# import pytest


@pytest.fixture(scope='module')
def source_logic_tree():
    config = Path(__file__).parent / 'fixtures' / 'source_logic_tree_sample_0.json'
    yield SourceLogicTree.from_json(config)


def test_slt_aliases(source_logic_tree):
    assert source_logic_tree.branch_sets[0] == source_logic_tree.fault_systems[0]

    assert source_logic_tree.fault_system_lts[0] == source_logic_tree.fault_systems[0]

    for filt_branch in source_logic_tree:
        assert filt_branch.fslt == filt_branch.branch_set
        assert filt_branch.slt == filt_branch.logic_tree
        break


def test_deprication_warning(source_logic_tree):
    with pytest.warns(DeprecationWarning):
        source_logic_tree.fault_systems

    with pytest.warns(DeprecationWarning):
        source_logic_tree.fault_system_lts

    with pytest.warns(DeprecationWarning):
        for filt_branch in source_logic_tree:
            filt_branch.fslt

    with pytest.warns(DeprecationWarning):
        for filt_branch in source_logic_tree:
            filt_branch.slt



#! python slt_config.py

"""
Functions for converting SLT config .py files as used in Runzi and THP etc into the standardised nzshm-model form.

NB: this is needed only for version 1 logic trees, Version 2 are defined directly in JSON.
"""

import copy
import importlib.util
import logging
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Union

from nzshm_model.logic_tree.source_logic_tree import BranchAttributeValue
from nzshm_model.logic_tree.source_logic_tree.version1 import (
    Branch,
    FaultSystemLogicTree,
    SourceLogicTree,
    SourceLogicTreeCorrelation,
)

log = logging.getLogger(__name__)

try:
    from .toshi_api import solution_rupt_set_id, toshi_api
except ModuleNotFoundError:
    log.warning("warning Toshi API module dependency not available, maybe you want to install with nzshm-model[toshi]")


def get_config_groups(logic_tree_permutations) -> Generator:
    for permutation in logic_tree_permutations[0][0]['permute']:
        yield permutation


def get_config_group(logic_tree_permutations: Iterable, group_tag: str) -> Union[Dict, None]:
    for group in get_config_groups(logic_tree_permutations):
        # print(group["group"], group_tag)
        if group['group'].upper() == group_tag.upper():
            return group
    return None


def get_config_group_tag_permutations(logic_tree_permutations: Iterable, group_tag: str) -> Generator:
    group = get_config_group(logic_tree_permutations, group_tag)
    if group:
        for member in group['members']:
            yield member['tag']


def common_tags(itm) -> Union[BranchAttributeValue, None]:
    try:
        if itm[0] == 'N':
            _n, _b = itm.split('^')
            return BranchAttributeValue(name='bN', long_name='bN pair', value=[float(_b[1:]), float(_n[1:])])
        if itm[0] == 'b':  # in some configs b is stand alone
            return BranchAttributeValue(name='b', long_name='b-value', value=float(itm[1:]))
        if itm[0] == 'C':
            return BranchAttributeValue(name='C', long_name='area-magnitude scaling', value=float(itm[1:]))
        if itm[0] == 's':
            return BranchAttributeValue(name='s', long_name='moment rate scaling', value=float(itm[1:]))
    except ValueError as err:
        print(err)
    return None


def decompose_subduction_tag(tag) -> Generator:
    """
    "tag": "Hik TL, N16.5, b0.95, C4, s0.42",
    "tag": "Puy 0.7, N4.6, b0.902, C4, s0.28",
    """
    if ", N" in tag and ", b" in tag:
        tag = tag.replace(", b", "^b")
    tag = tag.replace("Hik ", "Hik")
    tag = tag.replace("Puy ", "Puy")
    itms = tag.split(' ')

    for itm in itms:
        itm = itm.replace(',', '')  # remove commas

        if itm == 'hiktlck':
            yield BranchAttributeValue(name='dm', long_name='deformation model', value='TL')  # Trench Locked
            continue
        if itm == 'hiktcrp':
            yield BranchAttributeValue(name='dm', long_name='deformation model', value='TC')  # Trench Creeping
            continue
        if itm[:3].upper() in ["HIK", "PUY"]:
            yield BranchAttributeValue(name='dm', long_name='deformation model', value=itm[3:])
            continue

        other = common_tags(itm)
        if other:
            yield other


def decompose_crustal_tag(tag) -> Generator:
    """
    Examples:
     - "geodetic, TI, N2.7, b0.823 C4.2 s0.66"
     - "Cru_geol, b0.849, C4.1, s0.53"
    """
    tag = tag.replace("Cru_geol", "geologic")
    if ", N" in tag and ", b" in tag:
        tag = tag.replace(", b", "^b")
    itms = tag.split(' ')

    for itm in itms:
        itm = itm.replace(',', '')  # remove commas

        if "geo" in itm:
            yield BranchAttributeValue(name='dm', long_name='deformation model', value=itm)
            continue

        if itm in 'TI, TD':
            yield BranchAttributeValue(name='td', long_name='time dependent', value=(itm == "TD"))
            continue

        other = common_tags(itm)
        if other:
            yield other


def decompose_slab_tag(tag) -> Iterable:
    """used for SLAB. There is only one slab model in our logic tree"""
    return [
        BranchAttributeValue(name='r', long_name='earthquake rates', value='uniform'),
        BranchAttributeValue(name='d', long_name='hypocentral depths', value=1),
    ]


def from_config(config_path: Path, version: str = "", title: str = "") -> SourceLogicTree:
    """
    Build an SLT model from a config file, making some assumptions based on NSHM config conventions.
    """
    file_path = Path(config_path)
    module_name = "model"

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)  # type: ignore
    # sys.modules[module_name] = module
    spec.loader.exec_module(module)  # type: ignore

    def build_fslt(group_key, long_name):

        decompose_tag = None
        if group_key in ['PUY', 'HIK']:
            decompose_tag = decompose_subduction_tag
        if group_key == 'CRU':
            decompose_tag = decompose_crustal_tag
        if group_key == 'SLAB':
            decompose_tag = decompose_slab_tag

        group = get_config_group(module.logic_tree_permutations, group_key)

        if group:
            fslt = FaultSystemLogicTree(group_key, long_name)

            for member in group['members']:
                fslt.branches.append(
                    Branch(
                        values=list(decompose_tag(member['tag'])),
                        weight=member['weight'],
                        onfault_nrml_id=member['inv_id'],
                        distributed_nrml_id=member['bg_id'],
                    )
                )
            return fslt
        return

    fslts = [
        build_fslt(*group)
        for group in [("PUY", "Puysegur"), ("HIK", "Hikurangi-Kermadec"), ("CRU", "Crustal"), ("SLAB", "Intra-slab")]
    ]

    correlations = (
        build_correlations(module.src_correlations) if 'src_correlations' in dir(module) else []  # type: ignore
    )
    slt = SourceLogicTree(
        version=version,
        title=title,
        fault_system_lts=list(filter(lambda x: x is not None, fslts)),
        correlations=correlations,
    )
    return slt


def build_correlations(src_correlations: Dict[str, Any]) -> List[SourceLogicTreeCorrelation]:
    def select_decompose(group_name):
        if group_name.upper() in ['PUY', 'HIK']:
            return decompose_subduction_tag
        elif group_name.upper() == 'CRU':
            return decompose_crustal_tag
        elif group_name.upper() == 'SLAB':
            return decompose_slab_tag
        else:
            return None

    secondary_group = src_correlations['dropped_group']
    decompose_tag_sec = select_decompose(secondary_group)

    correlations = []
    for cor in src_correlations['correlations']:
        secondary_ind = [c['group'] for c in cor].index(secondary_group)
        primary_ind = int(not secondary_ind)
        decompose_tag_pri = select_decompose(cor[primary_ind]['group'])
        correlations.append(
            SourceLogicTreeCorrelation(
                primary_short_name=cor[primary_ind]['group'].upper(),
                secondary_short_name=secondary_group.upper(),
                primary_values=list(decompose_tag_pri(cor[primary_ind]['tag'])),
                secondary_values=list(decompose_tag_sec(cor[secondary_ind]['tag'])),
            )
        )

    return correlations


def resolve_toshi_source_ids(slt: SourceLogicTree) -> SourceLogicTree:
    new_slt = copy.deepcopy(slt)

    # SKIP_FS_NAMES =['HIK', 'CRU'] #, 'CRU'

    for fslt in new_slt.fault_system_lts:
        # if fslt.short_name in SKIP_FS_NAMES: #CRU
        #     continue
        if fslt:  # fslt can be None
            for branch in fslt.branches:
                nrml_info = toshi_api.get_source_from_nrml(branch.onfault_nrml_id)
                # print(nrml_info)
                branch.inversion_solution_id = nrml_info.solution_id
                branch.inversion_solution_type = nrml_info.typename
                if nrml_info.solution_id is not None:
                    branch.rupture_set_id = solution_rupt_set_id(nrml_info.solution_id)
                # print(branch) #, end='', flush=True)
                # print('.', end='', flush=True)
    return new_slt
